/*
 * FxGLUT version 0.9 - GLUT for Voodoo 1 and 2 under Linux
 * Copyright (C) 1999 Christopher John Purnell
 *                    cjp@lost.org.uk
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <termios.h>
#include <string.h>

#include "internal.h"
#include "menu.h"

#define MOUSE_BUFLEN 128

int __glutMouseFD = -1;
int __glutMouseX = 0;
int __glutMouseY = 0;

static int __glutMouseButtons = 0;

static void (*__glutMouseFunc)(int, int, int, int);
static void (*__glutMotionFunc)(int, int);
static void (*__glutPassiveMotionFunc)(int, int);

static void do_button(int, int);
static void do_movement(int, int);

static int parse_none(u_char *, int);
static int parse_ms(u_char *, int);
static int parse_ps2(u_char *, int);
static int parse_msc(u_char *, int);

static int (*parse_packet)(u_char *, int) = parse_ms;

#define C_NORM CREAD|CLOCAL|HUPCL

struct mouse_type
{
	char *name;
	u_short baud;
	u_short cflags;
	int (*func)(u_char *, int);
}
mouse_types[] =
{
	{ "ms", B1200, CS7|C_NORM, parse_ms },
	{ "ps2", 0, 0, parse_ps2 },
	{ "msc", B1200, CS8|CSTOPB|C_NORM, parse_msc },
	{ "none", 0, 0, parse_none },
	{ 0, 0, 0, 0 }
};

void __glutOpenMouse(void)
{
	struct termios tios;
	struct mouse_type *type = mouse_types;
	char *str;
	u_short baud;

	if (! (str = getenv("MOUSEDEV")) )
		str = "/dev/mouse";

	if ( (__glutMouseFD = open(str, O_RDONLY|O_NOCTTY|O_NONBLOCK)) < 0 )
		return;

	if ( (str = getenv("MOUSE")) )
	{
		struct mouse_type *ptr = mouse_types;

		while (ptr->name)
		{
			if (!strcmp(ptr->name, str))
			{
				type = ptr;
				break;
			}
			ptr++;
		}
	}

	parse_packet = type->func;

	if ( !(baud = type->baud) )
		return;

	if ( (str = getenv("MOUSEBAUD")) )
	{
		switch (atoi(str))
		{
		case 1200:
			baud = B1200;
			break;
		case 2400:
			baud = B2400;
			break;
		case 4800:
			baud = B4800;
			break;
		case 9600:
			baud = B9600;
			break;
		}
	}

	tcgetattr(__glutMouseFD, &tios);
	tios.c_cflag = baud | type->cflags;
	tios.c_iflag = IGNBRK;
	tios.c_oflag = 0;
	tios.c_lflag = 0;
	tios.c_cc[VMIN]  = 1;
	tios.c_cc[VTIME] = 0;
	tcsetattr(__glutMouseFD, TCSANOW, &tios);
}

void __glutCloseMouse(void)
{
	close(__glutMouseFD);
}

void __glutCheckMouse(void)
{
	static u_char buf[MOUSE_BUFLEN];
	static int len = 0;
	int got, pos;

	if (__glutMouseFD < 0)
		return;

	if ( (got = read(__glutMouseFD, buf + len, MOUSE_BUFLEN - len)) <= 0 )
		return;

	len += got;
	pos = 0;

	while ( (got = parse_packet(buf + pos, len)) )
	{
		pos += got;
		len -= got;
	}

	if (pos)
	{
		memcpy(buf, buf + pos, len);
	}
}

static int parse_none(u_char *buf, int len)
{
	return len;
}

static int parse_ms(u_char *buf, int len)
{
	static char state = 1;

	if (!state)
	{
		char dx, dy;

		if (len < 3)
			return 0;

		if ( !(buf[0] & 0x40) )
			return 1;

		dx = ((buf[0] & 0x03) << 6) | (buf[1] & 0x3F);
		dy = ((buf[0] & 0x0C) << 4) | (buf[2] & 0x3F);

		if (dx || dy)
			do_movement(dx, dy);

		do_button(GLUT_LEFT_BUTTON, buf[0] & 0x20);
		do_button(GLUT_RIGHT_BUTTON, buf[0] & 0x10);

		state = 1;
	}

	if (len < 4)
		return 0;

	state = 0;

	if ( (buf[3] & 0x40) )
		return 3;

	do_button(GLUT_MIDDLE_BUTTON, buf[3] & 0x20);

	return 4;
}

static int parse_ps2(u_char *buf, int len)
{
	int dx, dy;

	if (len < 3)
		return 0;

	if ((buf[0] & 0xC8) != 0x08)
		return 1;

	dx = (buf[0] & 0x10) ? buf[1] - 256 : buf[1];
	dy = (buf[0] & 0x20) ? -(buf[2] - 256) : -buf[2];

	if (dx || dy)
		do_movement(dx, dy);

	do_button(GLUT_LEFT_BUTTON, buf[0] & 0x01);
	do_button(GLUT_RIGHT_BUTTON, buf[0] & 0x02);
	do_button(GLUT_MIDDLE_BUTTON, buf[0] & 0x04);

	return 3;
}

static int parse_msc(u_char *buf, int len)
{
	int dx, dy;

	if (len < 5)
		return 0;

	if ( (buf[0] & 0xF8) != 0x80 )
		return 1;

	dx =  (char)buf[1] + (char)buf[3];
	dy = -(char)buf[2] - (char)buf[4];

	if (dx || dy)
		do_movement(dx, dy);

	do_button(GLUT_LEFT_BUTTON, ~buf[0] & 0x04);
	do_button(GLUT_RIGHT_BUTTON, ~buf[0] & 0x01);
	do_button(GLUT_MIDDLE_BUTTON, ~buf[0] & 0x2);

	return 5;
}

static void do_button(int button, int state)
{
	int mask = 1 << button;
	int old = __glutMouseButtons & mask;
	GlutMenu *menu;

	if (state)
	{
		__glutMouseButtons |= mask;

		if (!old)
		{
			if ( (menu = __glutAttachedMenus[button]) )
			{
				__glutMenuButtonDown(menu);
			}
			else if (__glutMouseFunc)
			{
				__glutMouseFunc(button, GLUT_DOWN,
						__glutMouseX, __glutMouseY);
			}
		}
	}
	else
	{
		__glutMouseButtons &= ~mask;

		if (old)
		{
			if ( (menu = __glutAttachedMenus[button]) )
			{
				__glutMenuButtonUp(menu);
			}
			else if (__glutMouseFunc)
			{
				__glutMouseFunc(button, GLUT_UP,
						__glutMouseX, __glutMouseY);
			}
		}
	}
}

static void do_movement(int x, int y)
{
	__glutMouseX += x;
	__glutMouseY += y;

	if (__glutMouseButtons)
	{
		if (__glutMotionFunc)
			__glutMotionFunc(__glutMouseX, __glutMouseY);
	}
	else
	{
		if (__glutPassiveMotionFunc)
			__glutPassiveMotionFunc(__glutMouseX, __glutMouseY);
	}

	if (__glutCurrentCursor >= 0 || __glutActiveMenu)
	{
		__glutPending |= GLUT_NEED_DISPLAY;
	}
}

void glutMouseFunc(void (*func)(int, int, int, int))
{
	__glutMouseFunc = func;
}

void glutMotionFunc(void (*func)(int, int))
{
	__glutMotionFunc = func;
}

void glutPassiveMotionFunc(void (*func)(int, int))
{
	__glutPassiveMotionFunc = func;
}

void glutWarpPointer(int x, int y)
{
	__glutMouseX = x;
	__glutMouseY = y;
}
